/*
    Sysepub 0.8.1 - A Symmetric Semi-Public key cipher
    Copyright (C) 2002 Pierre Blanc
    Pierre Blanc: blanc_teutoburgo@yahoo.it
    http://it.geocities.com/teutoburgo

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    or go to      http://www.gnu.org/copyleft/gpl.html
*/
import java.io.*;
import java.util.*;
/**
 * Sysepub 0.8.1 - A Symmetric Semi-Public key cipher
 * Creation date: (30/08/2002 16.50.29)
 * @author: Pierre Blanc
 */
public class Sysepub {
    public static final int MAX_DIM_BLOCKS_FACTOR=2;
    public static String SSP_HOME = "";
    private int period=1, defaultLength=3000;
    private A1timeP a1t;
    Syxor s = new Syxor();
    Random rnd = new Random();
    private long seed = 5345;
    private byte[] pattern;
    private String plain="plaintexts/", cipher="ciphertexts/",
	defaultFile="file";
    private String publicKey="publicKey/";
    private String patternString = "tEuToBuRgO";
    private int patternLength = patternString.length();
    /**
     * Sysepub constructor comment.
     */
    public Sysepub() {
	super();
	String temp = "";
	Properties prop = new Properties();
	FileInputStream fis=null, fis2=null;
	try {
	    fis = new FileInputStream(new File(SSP_HOME+"config/ssp.cfg"));
	} catch (FileNotFoundException fnfe){
	    try{
		fis2 = new FileInputStream(new File("c:/ssp_home.cfg"));
	    } catch (Exception e){
                System.out.println("Error! Can't find c:/ssp_home.cfg");
                System.out.println("Copy it from patch4win/ssp_home.cfg");
		System.exit(-1);
	    }
	    try{
		prop.load(fis2);
		SSP_HOME=prop.getProperty("SSP_HOME");
		fis = new FileInputStream(new File(SSP_HOME+"config/ssp.cfg"));
	    } catch (Exception e){
		e.printStackTrace();
		System.out.println("Using SSP_HOME="+SSP_HOME);
		System.out.println("Set the SSP_HOME");
		System.exit(-1);
	    }
	} catch (IOException ioe) {
	    ioe.printStackTrace();
	}
	try{
	    prop.load(fis);
	    temp = prop.getProperty("defaultfile","file");
	    defaultFile=temp;
	    temp = prop.getProperty("seed");
	    seed= Long.parseLong(temp);
	    patternString = prop.getProperty("patternstring");
		patternLength = patternString.length();
	    temp = prop.getProperty("period", "1");
	    period = Integer.parseInt(temp);
	    a1t = new A1timeP(period);
            temp = prop.getProperty("randomFileDefaultLength","3");
            defaultLength=Integer.parseInt(temp)*1024;
	} catch (NumberFormatException nfe) {
	    nfe.printStackTrace();
	} catch (IOException ioe){
	    ioe.printStackTrace();
	}

    }
    /**
     * Generates a random file in the publicKey directory
     * @param length the file length
     */
    public void generateRandomFile(int length){
      String fileName="publicKey";
      int fLength=fileName.length();
      int max=0;
      byte[] randomBytes = new byte[length];
      File f = new File(SSP_HOME+publicKey);
      File[] list = f.listFiles();
      try{
      for (int i=0; i<list.length; i++){
        if (!(fLength>list[i].getName().toString().length())){
          if(list[i].getName().toString().substring(0,fLength).equals(fileName))
          max++;
        }
      }
      }catch (Exception e){
        e.printStackTrace();
      }
      rnd.nextBytes(randomBytes);
      String progressive=Integer.toString(max+1);
      String generated=SSP_HOME+publicKey+fileName+progressive+".rnd";
      this.writeFile(generated,randomBytes);
      System.out.println("\nGenerated "+length+" random bytes in file "+generated);
    }
    /**
     * method for compacting the key files in a single array
     * @param mngs the array of the key files
     */
    public byte[] compactMngs(byte[][] mngs) {
	int i = 0, totalLength = 0, k = 0;
	for (i = 0; i < mngs.length; i++) {
	    totalLength = totalLength + mngs[i].length;
	}
	byte[] compacted = new byte[totalLength];
	for (i = 0; i < mngs.length; i++) {
	    for (int j = 0; j < mngs[i].length; j++) {
		compacted[k] = mngs[i][j];
		k++;
	    }
	}
	return compacted;
    }
    /**
     * The method that performs decryption
     * @param x3 The ciphertext array
     * @param pubKey The pubKey array
     * @return The x plaintext array (x1 cleaned from random blocks b1, b2)
     */
    public byte[] decryptFast(byte[] x3, byte[] pubKey) {
	int pubKeyL = pubKey.length;
	int l = x3.length;
	int i = 0, first = 0, last = 0, whereFound = 0;
	byte[] x1 = new byte[l];
	byte[] shifted = new byte[l];
	byte[] realKey = new byte[l];
	boolean notFound = false;
	byte[] x=null, prs;

	if (pattern == null)
        pattern = this.getPattern(patternString);
	System.out.println("Decrypting x3 ...");
        prs=a1t.getBytes(l, seed);
	byte[] x2 = s.decrypt(x3, prs);
	System.out.println("Decrypting x2 ...");
	int shift = a1t.getInt(seed, pubKeyL - l);
	//    System.out.println("---> shift=" + shift + " <-------");
	for (i = 0; i < l; i++) {
	    realKey[i] = pubKey[i + shift];
	}
	x1 = s.decrypt(x2, realKey);
	whereFound = this.findPattern(pattern, x1, 0);
	if (whereFound != -1) {
	    //        System.out.println("Pattern found");
	    first = whereFound;
	    whereFound = this.findPattern(pattern, x1, whereFound + 1);
	    if (whereFound != -1) {
		System.out.println("Ok.");
		last = whereFound;
		shift = l + pubKeyL - 2 * patternLength;
	    }
	}
	//    System.out.println("first " + first + " last " + last);
	if (last > 0) {
	    x = new byte[last - first - patternLength];
	    for (i = first + patternLength; i < last; i++) {
		x[i - first - patternLength] = x1[i];
	    }
	    System.out.println("Decryption performed.");
	} else {
	    System.out.println("Error - Can't decrypt");
	    System.exit(-1);
	}
	return x;
    }
    /**
     * Prepares for decryption and calls the decryptFast method <br>
     * Writes the plaintext array x in its directory
     * @param fileName The filename of the ciphertext
     */
    public void doDecryption(String fileName) {
	byte[] x;
	byte[] x3;

	System.out.println("Starting decryption of file " + SSP_HOME + cipher+fileName +
			   "  ...");
	byte[][] arrays = this.initArrays(cipher + fileName);
	x3 = arrays[0];
	byte[] pubKey = arrays[1];

	if (x3.length < pubKey.length)
	    x = this.decryptFast(x3, pubKey);

	else {
	    System.out.println("Error - The file is too big.. expand the"+
			       " publicKey files");
	    return;
	}
	this.writeFile(SSP_HOME + plain + fileName, x);
	System.out.println("Written "+SSP_HOME + plain+fileName);
    }
    /**
     * Prepares for encryption and calls the encryptFast method <br>
     * Writes the ciphertext array x3 in its directory
     * @param fileName The filename of the plaintext
     */
    public void doEncryption(String fileName) {
	int i, j, l, pr = 0, maxShift = 0, minShift = 0, cLength = 0;
	byte[] x, x3;

	System.out.println("Starting encryption of file "+SSP_HOME + plain+
                            fileName+"  ...");
	byte[][] arrays = this.initArrays(plain+fileName);
	x = arrays[0];

	byte[] pubKey = arrays[1];

	x3 = this.encryptFast(x, pubKey);

	if (x3.length > pubKey.length)
	    System.out.println("Error - The file is too big.. expand the publicKey"+
			       " directory");
		else{
			this.writeFile(SSP_HOME + cipher+fileName, x3);
			System.out.println("Written "+SSP_HOME + cipher+fileName);
		}
    }
    /**
     * The method that performs encryption
     * @param x The plaintext array
     * @param pubKey The pubKey array
     * @return The x3 ciphertext array
     */
    public byte[] encryptFast(byte[] x, byte[] pubKey) {
	int l = x.length;
	int pubKeyL = pubKey.length;
	int lTot = l + 2 * this.patternLength;
	int factor=MAX_DIM_BLOCKS_FACTOR;
	int n1 = rnd.nextInt(factor * (l + 2 * patternLength));
	int n2 = rnd.nextInt(factor * (l + 2 * patternLength));
	byte[] b1 = new byte[n1];
	byte[] b2 = new byte[n2];
	int x1Length = l + n1 + n2 + 2 * patternLength;
	byte[] x1 = new byte[x1Length];
	byte[] x3, x2, prs;
	int i = 0, pXLength = 2 * patternLength + l;

	pattern = this.getPattern(patternString);
	rnd.nextBytes(b1);
	rnd.nextBytes(b2);
	//*************************** appending the arrays ******
	for (i = 0; i < n1; i++) {
	    x1[i] = b1[i];
	}
	for (i = n1; i < n1 + patternLength; i++) {
	    x1[i] = pattern[i - n1];
	}
	for (i = n1 + patternLength; i < n1 + patternLength + l; i++) {
	    x1[i] = x[i - n1 - patternLength];
	}
	for (i = n1 + patternLength + l; i < n1 + pXLength; i++) {
	    x1[i] = pattern[i - (n1 + patternLength + l)];
	}
	for (i = n1 + pXLength; i < x1Length; i++) {
	    x1[i] = b2[i - (n1 + pXLength)];
	}
	//***********************************************************

	if (pubKeyL < x1Length) {
	    System.out.println("Error - publicKey files are too small");
	    return x;
	}
	byte[] realKey = new byte[x1Length];
	int maxShift = pubKeyL - x1Length;
	int pr = a1t.getInt(seed, maxShift);
	System.out.println("Encrypting x2 ...");
	for (i = 0; i < x1Length; i++) {
	    realKey[i] = pubKey[i + pr];
	}
	x2 = s.encrypt(x1, realKey);
	System.out.println("Encrypting x3 ...");
        prs=a1t.getBytes(x1Length, seed);
	x3 = s.encrypt(x2, prs);
	System.out.println("Encryption performed.");
	return x3;
    }
    /**
     * Searches for the pattern within x1
     * @param pattern The pattern array
     * @param x1 The x1 array (x concatenated with the random blocks b1,b2)
     * @param start The starting index within x1
     * @return The index of the pattern in x1
     */
    public int findPattern(byte[] pattern, byte[] x1, int start) {
	boolean notFound = false;
	int whereFound = -1;
	for (int i = start; i < x1.length - pattern.length; i++) {
	    for (int j = 0; j < pattern.length; j++) {
		if (pattern[j] != x1[i + j]) {
		    notFound = true;
		    j = pattern.length;
		} else
		    notFound = false;
	    }
	    if (!notFound) {
		whereFound = i;
		i = x1.length - pattern.length;
	    }
	}
	return whereFound;
    }
    public String getDefaultFile(){
	return defaultFile;
    }
    public int getDefaultLength(){
	return defaultLength;
    }
/*    public byte[] getPattern(long seed) {
	String p = (new Long(seed)).toString();
	seedLength = p.length();
	byte[] pattern = new byte[seedLength];
	for (int ss = 0; ss < seedLength; ss++) {
	    pattern[ss] = (byte) p.charAt(ss);
	}
	return pattern;
    }*/
    /**
     * Creates the pattern array from a String
     * @param patternString The pattern string
     * @return The pattern array
     */
    public byte[] getPattern(String patternString) {
	int pLength = patternString.length();
	byte[] pattern = new byte[pLength];
	for (int ss = 0; ss < pLength; ss++) {
	    pattern[ss] = (byte) patternString.charAt(ss);
	}
	return pattern;
    }
    /**
     * Makes the initialization of the arrays of plaintext/ciphertext and pubKey
     * @param pathname The pathName of the plaintext/ciphertext
     * @return Two arrays: fileToXOR the file to be XORed and pubKey the
     * public key
     */
    public byte[][] initArrays(String pathName) {
	byte[][] arrays=new byte[2][];
	byte[] fileToXOR;
	fileToXOR = this.readFile(SSP_HOME + pathName);
	byte[][] mngs = this.readPubKey(SSP_HOME + publicKey);
	byte[] pubKey = this.compactMngs(mngs);
	arrays[0]=fileToXOR;
	arrays[1]=pubKey;

	return arrays;
    }
    public static void main(String args[]) {
	System.out.println("Sysepub 0.8 Copyright (C) Pierre Blanc 2002"+
          "  http://it.geocities.com/teutoburgo");
        System.out.println("This software is under"+
          " the GNU GPL license v.2");
	Sysepub ssp = new Sysepub();
	int argsLen = args.length;
	String fileName = ssp.getDefaultFile();
        int length=ssp.getDefaultLength();
	String badParameter="Bad parameter.. type: java Sysepub -h";

	if (argsLen == 0) {
	    ssp.printHelp();
	} else
	    if (argsLen == 1) {
		if (args[0].equalsIgnoreCase("e"))
		    ssp.doEncryption(fileName);
		else
		    if (args[0].equalsIgnoreCase("d"))
			ssp.doDecryption(fileName);
		    else
			if (args[0].equalsIgnoreCase("h")
			    || args[0].equalsIgnoreCase("help")
			    || args[0].equalsIgnoreCase("-h")
			    || args[0].equalsIgnoreCase("--help"))
			    ssp.printHelp();
                        else if (args[0].equalsIgnoreCase("g"))
                          ssp.generateRandomFile(length);
			else if (args[0].equalsIgnoreCase("w")){
                          ssp.printNoWarranty();
                        }else
			    System.out.println(badParameter);
	    } else
		if (argsLen == 2)
		    if (args[0].equalsIgnoreCase("e"))
			ssp.doEncryption(args[1]);
		    else
			if (args[0].equalsIgnoreCase("d"))
			    ssp.doDecryption(args[1]);
                        else if (args[0].equalsIgnoreCase("g")){
                          try{
                            length=Integer.parseInt(args[1])*1024;
                          } catch (NumberFormatException nfe){
                            nfe.printStackTrace();
                            System.exit(-1);
                          }
                          ssp.generateRandomFile(length);
                        }
			else
			    System.out.println(badParameter);
		else
		    System.out.println("Wrong number of parameters");
    }
    public void printNoWarranty(){
      System.out.println("\n   BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY\n"+
"FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  EXCEPT WHEN\n"+
"OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES\n"+
"PROVIDE THE PROGRAM 'AS IS' WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED\n"+
"OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF\n"+
"MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS\n"+
"TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE\n"+
"PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING,\n"+
"REPAIR OR CORRECTION.\n\n"+

  " IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING\n"+
"WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR\n"+
"REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES,\n"+
"INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING\n"+
"OUT OF THE USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED\n"+
"TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY\n"+
"YOU OR THIRD PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER\n"+
"PROGRAMS), EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE\n"+
"POSSIBILITY OF SUCH DAMAGES.");
    }
    public void printHelp() {
	System.out.println("Sysepub comes with ABSOLUTELY NO WARRANTY; for details: java Sysepub w");
	System.out.println("This is free software, and you are welcome to redistribute it"+
	" under certain\nconditions; for details see the gpl.txt file.");
	System.out.println("Usage:\n");
	System.out.println("java Sysepub option [parameter]");
	System.out.println("");
	System.out.println("Options:");
	System.out.println("e : performs encryption (parameter: the filename to write)");
	System.out.println("d : performs decryption (parameter: the filename to write)");
	System.out.println("g : generates a random file as public key (parameter: the file length in KB)");
	System.out.println("If no filename is specified the default file is used.");
	System.out.println("Put the files in the proper directory; give only the filename, not the pathname");
	System.out.println();
	System.out.println("Examples: java Sysepub e supersecret.txt");
	System.out.println("Encrypts the file specified and writes it in the ciphertexts directory");
	System.out.println("          java Sysepub d supersecret.txt");
	System.out.println("Decrypts the file specified and writes it in the plaintexts directory");
        System.out.println("\nSee the readme.txt , doc/manual.html , doc/help.html "+
                          ", doc/gpl.txt for more\ninformations.");
//        System.out.println("There is a configuration file: config/ssp.cfg");
    }
    public byte[] readFile(String pathName) {
	File f = new File(pathName);
	long fLength = f.length();
	byte[] file = new byte[(int) fLength];

	try {
	    FileInputStream fis = new FileInputStream(f);
	    fis.read(file);
	} catch (IOException ioe) {
	    ioe.printStackTrace();
	    System.exit(-1);
	}

	return file;
    }
    /**
     * Reads the array pubKey from the publicKey directory
     * @param pathName The pathname
     * @return The array pubKey
     */
    public byte[][] readPubKey(String pathName) {
	File f = new File(pathName);
	File[] list = f.listFiles();
	int len = list.length, i = 0, j = 0;
	long fLength = 0;
	byte[][] mngs = new byte[len][];
	System.out.println("Reading "+len+" public key files:");
	for (i = 0; i < len; i++) {
	    System.out.println(list[i]);
	    if (list[i].isFile()) {
		fLength = list[i].length();
		try {
		    mngs[i] = new byte[(int) fLength];
		    (new FileInputStream(list[i])).read(mngs[i]);
		} catch (IOException ioe) {
		    ioe.printStackTrace();
                    System.exit(-1);
		}
		j++;
	    }
	}
	byte[][] pubKey = new byte[j][];
	for (i = 0; i < j; i++) {
	    pubKey[i] = mngs[i];
	}
	return pubKey;
    }
    public void writeFile(String name, byte[] bytes) {
	try {
	    FileOutputStream fos = new FileOutputStream(name);
	    fos.write(bytes);
	} catch (FileNotFoundException fnfe) {
	    fnfe.printStackTrace();
            System.exit(-1);
	} catch (IOException ioe){
	    ioe.printStackTrace();
            System.exit(-1);
	}
    }
}
